# Generated by Django 3.0.3 on 2020-02-13 06:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("datasets", "0002_add_datasettable"),
    ]

    operations = [
        migrations.AddField(
            model_name="datasettable",
            name="geometry_field_type",
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
    ]
